@@ -1,38 +1,36 @@
-import os
-import hashlib
-import base64
+from node.behaviors import Adopt
+from node.behaviors import Attributes
+from node.behaviors import DefaultInit
+from node.behaviors import NodeChildValidate
+from node.behaviors import Nodespaces
+from node.behaviors import Nodify
+from node.behaviors import OdictStorage
+from node.behaviors import Storage
+from node.compat import UNICODE_TYPE
+from node.ext.ugm import Group as BaseGroupBehavior
+from node.ext.ugm import Groups as BaseGroupsBehavior
+from node.ext.ugm import Ugm as BaseUgmBehavior
+from node.ext.ugm import User as BaseUserBehavior
+from node.ext.ugm import Users as BaseUsersBehavior
+from node.interfaces import IInvalidate
+from node.locking import locktree
+from node.locking import TreeLock
+from node.utils import UNSET
 from odict import odict
-from plumber import (
-    plumbing,
-    plumb,
-    default,
-    override,
-    Behavior,
-)
-from node.interfaces import IStorage
-from node.locking import (
-    locktree,
-    TreeLock,
-)
-from node.behaviors import (
-    NodeChildValidate,
-    Nodespaces,
-    Adopt,
-    Attributes,
-    DefaultInit,
-    Nodify,
-    Storage,
-    OdictStorage,
-)
-from node.ext.ugm import (
-    User as BaseUserBehavior,
-    Group as BaseGroupBehavior,
-    Users as BaseUsersBehavior,
-    Groups as BaseGroupsBehavior,
-    Ugm as BaseUgmBehavior,
-)
+from plumber import Behavior
+from plumber import default
+from plumber import override
+from plumber import plumbing
+from zope.interface import implementer
+import base64
+import hashlib
+import os
 
 
+ENCODING = 'utf-8'
+
+
+@implementer(IInvalidate)
 class FileStorage(Storage):
     """Storage behavior handling key/value pairs in a file.
 
@@ -42,8 +40,6 @@ class FileStorage(Storage):
          attribute.
     """
     allow_non_node_childs = override(True)
-    unicode_keys = default(True)
-    unicode_values = default(True)
     delimiter = default(':')
 
     @override
@@ -51,11 +47,12 @@ class FileStorage(Storage):
         self.__name__ = name
         self.__parent__ = parent
         self.file_path = file_path
+        self._storage_data = None
 
     @default
     @property
     def storage(self):
-        if not hasattr(self, '_storage_data'):
+        if self._storage_data is None:
             self._storage_data = odict()
             if self.file_path and os.path.isfile(self.file_path):
                 self.read_file()
@@ -64,33 +61,61 @@ class FileStorage(Storage):
     @default
     def read_file(self):
         data = self._storage_data
-        with open(self.file_path) as file:
-            for line in file:
-                k, v = line.split(self.delimiter)
-                if not isinstance(k, unicode) and self.unicode_keys:
-                    k = k.decode('utf-8')
-                if not isinstance(v, unicode) and self.unicode_values:
-                    v = v.decode('utf-8')
-                data[k] = v.strip('\n')
+        delimiter = self.delimiter.encode(ENCODING) \
+            if isinstance(self.delimiter, UNICODE_TYPE) \
+            else self.delimiter
+        with open(self.file_path, 'rb') as f:
+            for line in f:
+                idx = line.find(delimiter)
+                if idx == -1:
+                    # malformed line, ignore
+                    continue
+                k = line[:idx]
+                v = line[idx + len(delimiter):].strip(b'\n')
+                if v.startswith(b'b64:'):
+                    v = base64.b64decode(v[4:])
+                else:
+                    v = v.decode(ENCODING)
+                data[k.decode(ENCODING)] = v
 
     @default
     def write_file(self):
         lines = list()
-        if hasattr(self, '_storage_data'):
-            data = self._storage_data
-        else:
-            data = dict()
+        if self._storage_data is None:
+            if not os.path.exists(self.file_path):
+                with open(self.file_path, 'wb') as file:
+                    file.write(b'')
+            return
+        data = self._storage_data
+        delimiter = self.delimiter.encode(ENCODING) \
+            if isinstance(self.delimiter, UNICODE_TYPE) \
+            else self.delimiter
         for k, v in data.items():
-            if isinstance(k, unicode) and self.unicode_keys:
-                k = k.encode('utf-8')
-            if isinstance(v, unicode) and self.unicode_values:
-                v = v.encode('utf-8')
-            if v is None:
-                v = ''
-            line = self.delimiter.join([k, v]) + '\n'
+            if isinstance(k, UNICODE_TYPE):
+                k = k.encode(ENCODING)
+            if isinstance(v, UNICODE_TYPE):
+                v = v.encode(ENCODING)
+            elif v is None:
+                v = b''
+            elif v is UNSET:
+                v = b''
+            else:
+                v = b'b64:' + base64.b64encode(v)
+            line = delimiter.join([k, v]) + b'\n'
             lines.append(line)
-        with open(self.file_path, 'w') as file:
-            file.writelines(lines)
+        with open(self.file_path, 'wb') as f:
+            f.writelines(lines)
+
+    @override
+    def keys(self):
+        # Make pypy happy by overriding ``keys``
+        return self.storage.keys()
+
+    @override
+    def invalidate(self, key=None):
+        # This storage not provides invalidation by key. always entire storage
+        # gets invalidated
+        self._storage_data = None
 
     @default
     def __call__(self):
@@ -106,6 +131,20 @@ class FileAttributes(object):
     pass
 
 
+class UserAttributes(FileAttributes):
+
+    def __getitem__(self, key):
+        # provide id attribute expected by cone.ugm
+        # XXX: think about principal_id property on interface
+        if key == 'id':
+            return self.parent.name
+        # provide password attribute expected by cone.ugm
+        # XXX: think about password property on interface
+        if key == 'password':
+            return self.parent.parent.storage[self.parent.name]
+        return super(UserAttributes, self).__getitem__(key)
+
+
 class UserBehavior(BaseUserBehavior):
 
     @default
@@ -114,7 +153,7 @@ class UserBehavior(BaseUserBehavior):
         if not os.path.exists(user_data_dir):
             os.makedirs(user_data_dir)
         user_data_path = os.path.join(user_data_dir, parent.name)
-        return FileAttributes(name, parent, user_data_path)
+        return UserAttributes(name, parent, user_data_path)
 
     attributes_factory = default(user_data_attributes_factory)
 
@@ -129,6 +168,7 @@ class UserBehavior(BaseUserBehavior):
     def __call__(self, from_parent=False):
         self.attrs()
         if not from_parent:
+            self.parent()
             self.parent.parent.attrs()
 
     @default
@@ -175,6 +215,16 @@ class User(object):
     pass
 
 
+class GroupAttributes(FileAttributes):
+
+    def __getitem__(self, key):
+        # provide id attribute expected by cone.ugm
+        # XXX: think about principal_id property on interface
+        if key == 'id':
+            return self.parent.name
+        return super(GroupAttributes, self).__getitem__(key)
+
+
 class GroupBehavior(BaseGroupBehavior):
 
     @default
@@ -183,7 +233,7 @@ class GroupBehavior(BaseGroupBehavior):
         if not os.path.exists(group_data_dir):
             os.makedirs(group_data_dir)
         group_data_path = os.path.join(group_data_dir, parent.name)
-        return FileAttributes(name, parent, group_data_path)
+        return GroupAttributes(name, parent, group_data_path)
 
     attributes_factory = default(group_data_attributes_factory)
 
@@ -195,12 +245,14 @@ class GroupBehavior(BaseGroupBehavior):
 
     @default
     def __getitem__(self, key):
+        if key not in self.member_ids:
+            raise KeyError(key)
         return self.parent.parent.users[key]
 
     @default
     @locktree
     def __delitem__(self, key):
-        if not key in self.member_ids:
+        if key not in self.member_ids:
             raise KeyError(key)
         self._remove_member(key)
 
@@ -214,11 +266,13 @@ class GroupBehavior(BaseGroupBehavior):
     def __call__(self, from_parent=False):
         self.attrs()
         if not from_parent:
+            self.parent()
             self.parent.parent.attrs()
 
     @default
     def add(self, id):
-        if not id in self.member_ids:
+        if id not in self.member_ids:
+            self.parent.parent.users[id]
             self._add_member(id)
 
     @default
@@ -242,21 +296,21 @@ class GroupBehavior(BaseGroupBehavior):
     @default
     @property
     def member_ids(self):
-        return [id for id in self.parent.storage[self.name].split(',') if id]
+        return [id for id in self.parent.storage[self.name].split(u',') if id]
 
     @default
     def _add_member(self, id):
         member_ids = self.member_ids
         member_ids.append(id)
         member_ids = sorted(member_ids)
-        self.parent.storage[self.name] = ','.join(member_ids)
+        self.parent.storage[self.name] = u','.join(member_ids)
 
     @default
     def _remove_member(self, id):
         member_ids = self.member_ids
         member_ids.remove(id)
         member_ids = sorted(member_ids)
-        self.parent.storage[self.name] = ','.join(member_ids)
+        self.parent.storage[self.name] = u','.join(member_ids)
 
 
 @plumbing(
@@ -273,7 +327,7 @@ class SearchBehavior(Behavior):
 
     @default
     def _compare_value(self, term, value):
-        # XXX: this should be done by regular expressions.
+        # XXX: should this be done by regular expressions?
         if term == '*':
             return True
         if not len(term.strip('*')):
@@ -301,20 +355,23 @@ class SearchBehavior(Behavior):
         setups with just a few users and groups.
         """
         found = set()
-        if criteria is None:
-            return list()
         for principal in self.values():
             # exact match too many
             if exact_match and len(found) > 1:
-                raise ValueError(
-                    u"Exact match asked but result not unique")
+                raise ValueError('Exact match asked but result not unique')
+            # no criteria, all principals match
+            if not criteria:
+                found.add(principal)
+                continue
             # or search
             if or_search:
                 for key, term in criteria.items():
                     if key == 'id':
                         if self._compare_value(term, principal.name):
                             found.add(principal)
-                        continue
+                        # continue never executed due to cpython peephole
+                        # optimization, thus not counted in coverage
+                        continue                             # pragma: no cover
                     value = principal.attrs.get(key)
                     if value:
                         if self._compare_value(term, value):
@@ -328,7 +385,9 @@ class SearchBehavior(Behavior):
                         if not self._compare_value(term, principal.name):
                             matches = False
                             break
-                        continue
+                        # continue never executed due to cpython peephole
+                        # optimization, thus not counted in coverage
+                        continue                             # pragma: no cover
                     value = principal.attrs.get(key)
                     if not value or not self._compare_value(term, value):
                         matches = False
@@ -337,7 +396,7 @@ class SearchBehavior(Behavior):
                     found.add(principal)
         # exact match zero found
         if exact_match and len(found) == 0:
-            raise ValueError(u"Exact match asked but result length is zero")
+            raise ValueError('Exact match asked but result length is zero')
         # attr list
         if attrlist:
             ret = list()
@@ -356,8 +415,6 @@ class SearchBehavior(Behavior):
 
 
 class UsersBehavior(SearchBehavior, BaseUsersBehavior):
-
-    unicode_values = default(False)
     salt_len = default(8)
     hash_func = default(hashlib.sha256)
 
@@ -368,26 +425,31 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         self.__parent__ = parent
         self.file_path = file_path
         self.data_directory = data_directory
+        self._storage_data = None
         self._mem_storage = dict()
+        self._user_data_to_remove = list()
 
     @override
     def __getitem__(self, key):
-        if not key in self.storage:
-            raise KeyError(key)
-        if key in self._mem_storage:
+        # access storage, if key not contained, KeyError is raised
+        self.storage[key]
+        try:
             return self._mem_storage[key]
-        with TreeLock(self):
-            user = User(
-                name=key, parent=self, data_directory=self.data_directory)
-            self._mem_storage[key] = user
-            return user
+        except KeyError:
+            with TreeLock(self):
+                user = self._mem_storage[key] = User(
+                    name=key,
+                    parent=self,
+                    data_directory=self.data_directory
+                )
+                return user
 
     @override
     @locktree
     def __setitem__(self, key, value):
         # set empty password on new added user.
-        if not key in self.storage:
-            self.storage[key] = ''
+        if key not in self.storage:
+            self.storage[key] = u''
         self._mem_storage[key] = value
 
     @override
@@ -400,15 +462,22 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         del self._mem_storage[key]
         if key in self.parent.attrs:
             del self.parent.attrs[key]
+        self._user_data_to_remove.append(key)
 
     @override
     @locktree
     def __call__(self, from_parent=False):
         self.write_file()
         for value in self.values():
-            value(True)
+            value(from_parent=True)
         if not from_parent:
             self.parent.attrs()
+            self.parent.groups(from_parent=True)
+        for userid in self._user_data_to_remove:
+            user_data_path = os.path.join(self.data_directory, 'users', userid)
+            if os.path.exists(user_data_path):
+                os.remove(user_data_path)
+        self._user_data_to_remove = list()
 
     @default
     def create(self, id, **kw):
@@ -425,7 +494,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
 
     @default
     def authenticate(self, id=None, pw=None):
-        if not id in self.storage:
+        if id not in self.storage:
             return False
         # cannot authenticate user with unset password
         if not self.storage[id]:
@@ -434,22 +503,27 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
 
     @default
     def passwd(self, id, oldpw, newpw):
-        if not id in self.storage:
-            raise ValueError(u"User with id '%s' does not exist." % id)
-        # case pwd overwrite
-        if self.storage[id]:
+        if id not in self.storage:
+            raise ValueError(u"User with id '{}' does not exist.".format(id))
+        if oldpw is not None:
             if not self._chk_pw(oldpw, self.storage[id]):
-                raise ValueError(u"Old password does not match.")
+                raise ValueError('Old password does not match.')
         salt = os.urandom(self.salt_len)
+        newpw = newpw.encode(ENCODING) \
+            if isinstance(newpw, UNICODE_TYPE) \
+            else newpw
         hashed = base64.b64encode(self.hash_func(newpw + salt).digest() + salt)
-        self.storage[id] = hashed
+        self.storage[id] = hashed.decode()
+        self()
 
     @default
     def _chk_pw(self, plain, hashed):
         hashed = base64.b64decode(hashed)
         salt = hashed[-self.salt_len:]
-        return hashed == self.hash_func(
-            plain.encode('utf-8') + salt).digest() + salt
+        plain = plain.encode(ENCODING) \
+            if isinstance(plain, UNICODE_TYPE) \
+            else plain
+        return hashed == self.hash_func(plain + salt).digest() + salt
 
 
 @plumbing(
@@ -473,26 +547,31 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
         self.__parent__ = parent
         self.file_path = file_path
         self.data_directory = data_directory
+        self._storage_data = None
         self._mem_storage = dict()
+        self._group_data_to_remove = list()
 
     @override
     def __getitem__(self, key):
-        if not key in self.storage:
-            raise KeyError(key)
-        if key in self._mem_storage:
+        # access storage, if key not contained, KeyError is raised
+        self.storage[key]
+        try:
             return self._mem_storage[key]
-        with TreeLock(self):
-            group = Group(
-                name=key, parent=self, data_directory=self.data_directory)
-            self._mem_storage[key] = group
-            return group
+        except KeyError:
+            with TreeLock(self):
+                group = self._mem_storage[key] = Group(
+                    name=key,
+                    parent=self,
+                    data_directory=self.data_directory
+                )
+                return group
 
     @override
     @locktree
     def __setitem__(self, key, value):
         # set empty group members on new added group.
-        if not key in self.storage:
-            self.storage[key] = ''
+        if key not in self.storage:
+            self.storage[key] = u''
         self._mem_storage[key] = value
 
     @override
@@ -501,18 +580,26 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
         del self.storage[key]
         if key in self._mem_storage:
             del self._mem_storage[key]
-        id = 'group:%s' % key
+        id = 'group:{}'.format(key)
         if id in self.parent.attrs:
             del self.parent.attrs[id]
+        self._group_data_to_remove.append(key)
 
     @override
     @locktree
     def __call__(self, from_parent=False):
         self.write_file()
         for value in self.values():
-            value(True)
+            value(from_parent=True)
         if not from_parent:
             self.parent.attrs()
+            self.parent.users(from_parent=True)
+        for groupid in self._group_data_to_remove:
+            group_data_path = os.path.join(
+                self.data_directory, 'groups', groupid)
+            if os.path.exists(group_data_path):
+                os.remove(group_data_path)
+        self._group_data_to_remove = list()
 
     @default
     def create(self, id, **kw):
@@ -562,13 +649,17 @@ class UgmBehavior(BaseUgmBehavior):
 
     @override
     def __getitem__(self, key):
-        if not key in self.storage:
+        if key not in self.storage:
             if key == 'users':
-                self['users'] = Users(file_path=self.users_file,
-                                      data_directory=self.data_directory)
+                self['users'] = Users(
+                    file_path=self.users_file,
+                    data_directory=self.data_directory
+                )
             else:
-                self['groups'] = Groups(file_path=self.groups_file,
-                                      data_directory=self.data_directory)
+                self['groups'] = Groups(
+                    file_path=self.groups_file,
+                    data_directory=self.data_directory
+                )
         return self.storage[key]
 
     @override
@@ -579,7 +670,7 @@ class UgmBehavior(BaseUgmBehavior):
 
     @override
     def __delitem__(self, key):
-        raise NotImplementedError(u"Operation forbidden on this node.")
+        raise NotImplementedError('Operation forbidden on this node.')
 
     @override
     def __iter__(self):
@@ -590,8 +681,8 @@ class UgmBehavior(BaseUgmBehavior):
     @locktree
     def __call__(self):
         self.attrs()
-        self.users(True)
-        self.groups(True)
+        self.users(from_parent=True)
+        self.groups(from_parent=True)
 
     @default
     @property
@@ -618,38 +709,38 @@ class UgmBehavior(BaseUgmBehavior):
     def add_role(self, role, principal):
         roles = self.roles(principal)
         if role in roles:
-            raise ValueError(u"Principal already has role '%s'" % role)
+            raise ValueError(u"Principal already has role '{}'".format(role))
         roles.append(role)
         roles = sorted(roles)
-        self.attrs[self._principal_id(principal)] = ','.join(roles)
+        self.attrs[self._principal_id(principal)] = u','.join(roles)
 
     @default
     @locktree
     def remove_role(self, role, principal):
         roles = self.roles(principal)
-        if not role in roles:
-            raise ValueError(u"Principal does not has role '%s'" % role)
+        if role not in roles:
+            raise ValueError(u"Principal does not has role '{}'".format(role))
         roles.remove(role)
         roles = sorted(roles)
-        self.attrs[self._principal_id(principal)] = ','.join(roles)
+        self.attrs[self._principal_id(principal)] = u','.join(roles)
 
     @default
     def _principal_id(self, principal):
         id = principal.name
         if isinstance(principal, Group):
-            id = 'group:%s' % id
+            id = 'group:{}'.format(id)
         return id
 
     @default
     def _roles(self, id):
         attrs = self.attrs
-        if not id in attrs:
+        if id not in attrs:
             return list()
         return [role for role in attrs[id].split(',') if role]
 
     @default
     def _chk_key(self, key):
-        if not key in ['users', 'groups']:
+        if key not in ['users', 'groups']:
             raise KeyError(key)
 
 
@@ -663,4 +754,9 @@ class UgmBehavior(BaseUgmBehavior):
     Nodify,
     OdictStorage)
 class Ugm(object):
-    pass
+
+    def invalidate(self, key=None):
+        if key is None:
+            self.storage.clear()
+            return
+        del self.storage[key]
