@@ -1,8 +1,7 @@
+from node.interfaces import IInvalidate
+from node.interfaces import ILeaf
+from node.interfaces import INode
 from zope.interface import Attribute
-from node.interfaces import (
-    INode,
-    ILeaf,
-)
 
 
 class IPrincipal(INode):
@@ -45,8 +44,8 @@ class IUser(IPrincipal, ILeaf):
         """Set password for this user.
 
         oldpw
-            Old password. If None, newpassword is always written. This
-            is needed for adding new users.
+            Old password. If old password is None, new password is always
+            set, otherwise an Exception is raised.
 
         newpw
             New password.
@@ -68,14 +67,34 @@ class IGroup(IPrincipal):
     member_ids = Attribute(u"List of member id's contained in this group.")
 
     def add(id):
-        """Add user with id to group.
+        """Add user with id to group. Raise ``KeyError`` if user not exists.
 
         id
             User id.
         """
 
+    def __getitem__(key):
+        """Return group related ``User`` object. Raise ``KeyError`` if user not
+        member of this group.
 
-class IPrincipals(INode):
+        key
+            User id.
+        """
+
+    def __delitem__(key):
+        """Delete membership of user from this group. Raise ``KeyError`` if user not
+        member of this group.
+
+        key
+            User id.
+        """
+
+    def __iter__():
+        """Iterate over ``member_ids``.
+        """
+
+
+class IPrincipals(INode, IInvalidate):
     """Interface describing a principals container.
     """
 
@@ -86,13 +105,14 @@ class IPrincipals(INode):
         """Search for principals.
 
         criteria
-            Dict like object defining the principal attributes to be matched. 
+            Dict like object defining the principal attributes to be matched.
+            If no criteria given, all users are returned.
 
         attrlist
-            Normally a list of keys is returned. By defining attrlist the
-            return format will be ``[(key, {attr1: [value1, ...]}), ...]``. To
+            if no ``attrlist`` is given a list of ids is returned. By defining attrlist the
+            return format will be ``[(id, {attr1: value1, attr2: value2, ...}), ...]``. To
             get this format without any attributs, i.e. empty dicts in the
-            tuples, specify an empty attrlist. 
+            tuples, specify an empty attrlist.
 
         exact_match
             raise ValueError if not one match, return format is a single key or
@@ -100,6 +120,8 @@ class IPrincipals(INode):
 
         or_search
             flag whether criteria should be ORer or ANDed. defaults to False.
+
+        :return: list of ids - if no ``attrlist`` is given else see ``attrlist``
         """
 
     def create(_id, **kw):
@@ -158,7 +180,7 @@ class IGroups(IPrincipals):
     """
 
 
-class IUgm(INode):
+class IUgm(INode, IInvalidate):
     """Interface describing user and group management API.
     """
 
@@ -166,6 +188,8 @@ class IUgm(INode):
 
     groups = Attribute(u"IGroups implementation")
 
+    # XXX: roles storage is implementation specific and should be removed from
+    #      here
     roles_storage = Attribute(u"Callable for persisting roles.")
 
     def add_role(role, principal):
